
/* CBG slider 68 series ASSEMBLER
 *
 *
 *
 * Tenison Tech Motorola 68K Assembler
 * (C) 1990 TenisonTech
 * This version not for commercial use.
 * License available from info@tenisontech.com
 *
 * No liability is accepted any errors in this material or for any loss.
 *
 */

 
char *assembler = "68k";
 
#include <stdio.h>
#include "a6hdr.h"
#include <malloc.h>
#include <ctype.h>
#include <strings.h>
 
extern void lok();       // Two entry points to 68k specific code.
extern void initspecific();

#define  OBJMAX  500 /* ?? */
#define  labmax  25  /* Max label length */
#define  linemax 400 /* longest input line */


void objword();
void objbyte();
int eval3();


char lable[OBJMAX], inst[OBJMAX], token[OBJMAX], etoken[OBJMAX];
int cards;        /* Total lines read */
int line  ;      /* line number */
char *currentsrc, *oldsrc;

int baddr[2]  ;      /* object address vector */
int seclen[2];       /* length of text and bss sections */
int object[OBJMAX];      /* bcpl string containing object bytes */
int pass;        /* 1 or 2 */
char *jobname;   /* string containing filename */
int eos, running;/* boolean flags end of src */
int errors    ;  /* error counter */
char *error_type;  /* zero for Ok */
int tab       ;  /* index for inst array */
int tvalue    ;  /* type of instruction */
char ech       ;  /* as ch but for eval */
int imed      ;  /* boolean set by # sign */
int oldline;     /* line in primary src file */
char * gnem   ; /* Current nemonic info line */
int readonlyopt; /* Set if object file must not update itself */

struct stnode *wordnode, **spoi, **stab_start, *stp;


FILE  *srccis, *oldcis, *objcis, *listcis;

union
{
  int i; char *s;
} error_sup;  /* Additional error information */

int xaddr     ;  /* memory allocation skip */
int error_tab ;  /* point of error */
int equate    ;  /* boolean flag for address print */

int relorigin ;  /* boolean flag for relocatable code */
int equateval ;  /* Value assigned in an equate */
int output = O_TEXT;  /* Type of code being generated */

void error(char *e, int s)
{ 
  if (error_type == NULL)
    {
      error_type = e;
      error_tab = tab;
      error_sup.i = s;
    }
}

void error_s(char *e, char *s)
{ 
  if (error_type == NULL)
    {
      error_type = e;
      error_tab = tab;
      error_sup.s = s;
    }
}

char *inpoi;
char inbuffer[linemax];

char mrdch(short x)
{
  char c = *(inpoi++);
  if (c == 9) c = ' ';
  if (x) c = toupper(c);
  return c;
}


void readline() 
 {  int i;
    char ch;
    if (feof(srccis))
        {  eos = 1;
           return;
        }
    fgets(inbuffer, linemax, srccis);
    inpoi = inbuffer;
    ch = mrdch(1);
    cards ++;
    line = line + 1;
    lable[0] = (char) 0;
    inst[0] = (char) 0;
    if (ch == '\n')  
        {  return;
        }

/* read lable */
    if (ch !=' ' && ch != ';' && ch != '\n' && ch != '*')
     {  i = 0;
        while (ch !=' ' && ch != '\n'  && ch != EOF)
         {  lable[i++] = ch;
            ch = mrdch(1);           // read label in capitals
            if (i >= labmax)  
                {  error("label too long", 0);
                   break;
                }
          }
        lable[i++] = (char) 0;
      }

    while (ch==' ')  ch = mrdch(1);  /* read instruction or comment */

    i = 0;
    while (ch !='\n' && ch != EOF)
     {  inst[i++] = ch;
        ch = mrdch(0);
        if (i > OBJMAX)  break;
      }
    inst[i++] = (char) 0;

  }

void newline() 
{ 
  printf("\n");
}

void writeline(FILE *ss) 
{
  int i;
  fprintf(ss, "%4i ", line);
  
  if (lable[0] ==';')
    { fprintf(ss, "%s %s\n", lable, inst);
    return;
    }
  
  if (object[0] > 0)  
    { 
      int j = (object[0] > 4) ? 4 : object[0];
      fprintf(ss, "%06X ", baddr[output]);
      
      for (i=1; i<=j; i++) fprintf(ss, "%02X", object[i]);
      for (i=j; i<=4; i++) fprintf(ss, "  ");
      fprintf(ss, " ");
    }
  else if (equate)  fprintf(ss, "=%08X         ", equateval);
  else fprintf(ss, "              ");
  
  fprintf(ss, "%25s", lable);
  fprintf(ss, "  %s\n", inst);
  
  if (object[0] > 4)
    {
      fprintf(ss, "            ");
      for (i=5; i <= object[0]; i++)
	{
	  fprintf(ss, "%02X", object[i]);
	}
      fprintf(ss, "\n");
    }
}


int hash(char *s)
{ 
  int r=0;
  while(*s)
  {
    r = r + r + *s;
    s++;
  }
  return r & (HASHSIZE-1);
}

int lookup(char *lab)               /* search symbol table */
{  
  wordnode = stab_start[hash(lab)];
  while (wordnode)
    { 
      /* printf("<%s> <%s> %i %i \n", lab, wordnode->textual, hash(lab), hash(wordnode->textual));  */
      
      if (strcmp(lab, wordnode->textual)==0)  return 1;
      wordnode = wordnode->stnext;
    }
  return 0;
}

void say_error(FILE *ss)
{ 
  int i;
  errors = errors + 1;
  fprintf(ss, "ERROR--------------------------------------");
  for (i=1; i<= error_tab; i++)  putc('-', ss);
  fprintf(ss, "^\n'%s': ", currentsrc);
  fprintf(ss, error_type, error_sup);
  fprintf(ss, "\n");
}

char nextch(skip, upper) 
{
  ech = inst[tab++];
  if (skip && (ech == ' '))  nextch(1, 1);       /* skip leading spaces (recursively) */
  if (upper) ech = toupper(ech);
  return ech;
}

void ignblank()
{
  while (ech == ' ') nextch(1, 1);
}

void chcomma() 
{
  if (ech!=',')  error("Comma expected.", 0);
  nextch(1, 1);
}

void checkfor(char c)
{
  if (ech != c)  error("<%c> expected.", c);
  nextch(1, 1);
}

int aschex(char cch)
{ 
  int ch = cch;
  return  (ch < '0') ? -1:
    (ch > 'F') ? -1:
    (ch <= '9') ? ch - '0':
    (ch >= 'A') ? ch - 'A' + 10: -1;
}

int alnum(char ch)
{ 
  return (ch == '_') ? 1:
    (ch == '.') ? 1:
    (ch < '0') ? 0:
    (ch > 'Z') ? 0:
    (ch <= '9') ? 1:
    (ch >= 'A') ? 1: 0;
}



/* Recursive descent expression evaluation */
int eval9(dummy) 
{
  /*//if (ech == '#')
  //    { nextch(1, 1);
  //      if (ech == 'X') ech = '$';
  //      else error("Bad number base.");
  //    }
  */
  if (ech == '$')                 /* HEX CONSTANT */
    { 
      int s=0;
      nextch(1, 1);
      while (aschex(ech) >= 0)
	{  s = (s<<4) + aschex(ech);
	nextch(1, 1);
	}
      return s;
    }
  
  if (isdigit(ech))                  /* decimal constant */
    { 
      int s = 0; int base = 10;
      while (aschex(ech) >= 0)
	{  
	  s = s*base + aschex(ech);
	  nextch(1, 1);
	  if (ech == 'X') { base = 16; nextch(1, 1); }
	}
      return s;
    }
  
  if (ech == '-')                 /* monadic negate */
    {
      nextch(1, 1);
      return 0-eval9(tab);
    }
  
  if (ech == '~')                 /* one's complement */
    {
      nextch(1, 1);
      return -1-eval9(tab);
    }

  if (ech == '*')   /*assembly point */
    {
      nextch(1, 1);
      return baddr[output];
    }

    if (ech == '(')                /* expression in parens */
      {
	int r;
        nextch(1, 1);
        r = eval3();
        checkfor(')');
        return r;
      }

    if (ech == '\'')                /* ascii quote char */
      { 
	int r = nextch(0, 0);
        if (nextch(1, 1) != '\'')  error("Mismatched quotes.", 0);
        nextch(1, 1);
        return r;
      }
    
    
    if (ech== '#' || ech ==  '=')       /* immediate addressing flag */
      {
	int js;
	nextch(1, 1);
        js = eval9(tab);
        imed = 1;
        return js;
      }
    
    if ((isupper(ech)) || (ech=='_') )       /* symbol look up */
      {
	int i = 0;
        while (alnum(ech))
	  {
	    token[i++] = ech;
	    nextch(0, 1);
	  }
        token[i++] = (char) 0;
        ignblank();
        if (lookup(token))
	  { 
	    int j = wordnode->type;   /* symbol found */
	    if (j==RELOCATABLE || j==EXTERN)
	      {
		if (* spoi) error("Doublely relocation", 0);
		*spoi = wordnode;
	      }
	    if (j == EXTERN) return 0;   /* Do not give BSS offsets - ld adds them */
	    return wordnode->value;      /* return value */
          }
        else
	  { 
	    strcpy(etoken, token);
	    error_s("Symbol <%s> not defined.", etoken);
	    return 0;
	  }
      }
    error("Unknown thing in expression.", 0);
    return 0;
}


int eval5() 
{ 
  return  eval9(tab);
}

int eval4() 
{
  int s = eval5();
  while (ech == '*')  
     {  nextch(1, 1) ;  s = s * eval5();  }
  return s;
 }


int eval3() 
{
  int s = eval4();
  while (ech == '+' || ech == '-')  
  { 
    char k = ech;
    nextch(1, 1); 
    if (k == '+') s = s + eval4();  
    else s = s - eval4();  
  }
  return s;
}

/* Check data size */
int sexmeasure(data) 
{
  int k = (data  >= 0) ? data: -1- data;
  return (k<128) ? 0:              /* byte result */
    (k<32768) ? 1:           /* word */
    2;                       /* long word */
}


int eval(struct stnode **lspoi)
{  
  int res;
  spoi = lspoi;
  *spoi = NULL;
  res = eval3();
  return res;
}

/* IF big endian */
void objword(w)
{ 
  objbyte(w >> 8);
  objbyte(w);
}

void objbyte(b)
{
  int x = object[0];
  object[++x] = b & 0xFF;
  object[0] = x;
}


/* Note that an external or relocatable symbol has been used */
void symbnote(struct stnode *wn, int xvalue) 
{ 
  struct locchain *r;
  if (pass!=2) return;
  r = (struct locchain *) malloc(sizeof(struct locchain));
  r->next = wn->schain;
  wn->schain = r;
  r->val = xvalue;
}


/* put a lable into symbol table */
void lable_handle(char sf)  
{
  if (lookup(lable))
    { 
      if (pass==1 && sf != 'e') 
      {
	if (wordnode->type == EXTERN)   /* extern actually defined here */
	  { 
	    if (output == O_TEXT)
	      { 
		wordnode->value = baddr[O_TEXT];
		wordnode->type = (relorigin) ? RELOCATABLE: ABSOLUTE;
	      }
	    else /* Convert extern to local bss */
	      {
		wordnode->value = baddr[O_BSS];
		wordnode->equatef = EXTERN;
		wordnode->type = EXTERN;
		
		if (objcis) fprintf(objcis, "-B %s %08X\n",  lable, baddr[O_BSS]);
	      }
	  }
	else wordnode-> type  = DMTO;   /* found twice on first pass, dmto */
      }
      else if (pass == 2)
	{
	  if (wordnode->type==DMTO)  error("Defined more than once", 0);
	  if (wordnode->value != baddr[output] && wordnode->equatef == 0)
	    {
	      printf("Label value changed for %s\n", lable);
	      printf("type = %i, pass =%i, equatef=%i\n", wordnode->type, pass,
		     wordnode->equatef);
	    }
	}
    }
  else /* Not found before */
  { 
    int h = hash(lable);
    wordnode = (struct stnode *) malloc(sizeof(struct stnode)+4);
    wordnode->stnext = stab_start[h];
    stab_start[h] = wordnode;
    if (sf == 'e') 
      {
	wordnode->value = 0;
	wordnode->type = EXTERN;
	wordnode->equatef = EXTERN;
      }    
    else if (output == O_TEXT)
      {
	wordnode->value = baddr[O_TEXT];         /* initially set up as code lable */
	wordnode->equatef = 0;           /* So clear equate flag           */
	wordnode->type = (relorigin) ? RELOCATABLE: ABSOLUTE;
      }
    else if (output == O_BSS)
      {
	/* Enter bss labels as externs in symbol table */
	wordnode->value = baddr[O_BSS];
	wordnode->equatef = EXTERN;
	wordnode->type = EXTERN;
	if (objcis) fprintf(objcis, "-B %s %08X\n",  lable, baddr[O_BSS]);
      }
    
    wordnode->schain = NULL;
    wordnode->textual = (char *) malloc(strlen(lable)+1);
    strcpy(wordnode->textual, lable);
  }
}

void objhex(unsigned char c)
{
  char c1 = c >> 4;
  if (c1 > 9) c1 += 7;
  putc(c1 + '0', objcis);
  c1 = c & 15;
  if (c1 > 9) c1 += 7;
  putc(c1 + '0', objcis);
}

void code_output() 
{
  int i, sum = 0;
  if (objcis == NULL) return;
  
  if (relorigin && xaddr && output == O_TEXT)  /* Fill uninitialised storage */
    {
      
      for (i=0; i< xaddr; i++)  fprintf(objcis, "00");
    }
  
  if (object[0] == 0) return;
  
  if (output != O_TEXT) error("code in bss", 0);
  if (relorigin==0)
    { 
      fprintf(objcis, ":%02X%04X00", object[0], baddr[output]);
      sum = sum + object[0] + (baddr[output] & 255) + ((baddr[output] >> 8) & 255);
    }
  for (i=1; i<=object[0]; i++) 
    {
      objhex(object[i]);
      sum = object[i] + sum;
    }
  if (relorigin == 0) fprintf(objcis, "%02X\n", (256 - sum) & 255);
  fprintf(objcis, "\n");
}



void unstackcis()
{
  if (oldcis == NULL) running = 0;
  else
    { 
      srccis = oldcis;
      oldcis = NULL;
      line = oldline;
      eos = 0;
      currentsrc = oldsrc;
      readline();
    }
}

void perform_pass()
{ 
  char sf; running = 1;
  output = O_TEXT;
  eos = 0;
  cards = 0;
  seclen[0] = 0; /* No output text yet */
  seclen[1] = 0; /* No output text yet */
  /*  printf("Source: '%s' pass %i.\n", jobname, pass);
   */
  currentsrc = jobname;
  srccis = fopen(jobname, "r");
  if (srccis == NULL) 
    {
      printf("Can't open src file '%s' \n", jobname);
      perror(jobname);
      exit(1);
    }
  line = 0;
  relorigin = 1;  /* default to relocatable module */
  baddr[O_TEXT] = 0;          /* default origin */
  baddr[O_BSS] = 0;          /* default origin */
  errors = 0;
  
  while (running)
    {
      error_type = NULL;
      object[0] = 0; 
      readline();
      if (eos) unstackcis();
      if (running == 0) break;
      equate = 0;
      xaddr = 0;
      if (lable[0] == ';') continue;
      sf = (char) 0;
      
      if (strncmp(inst, "Extern", 6)==0) sf = 'e';
      if (strncmp(inst, "Global", 6)==0) sf = 'g';
      
      if (lable[0] != (char) 0) lable_handle(sf);
      else wordnode = NULL;
      
      tab = 0;
      lok(); 
      if ((error_type !=NULL) && (pass==2) && (listcis == NULL))
        {  
          writeline(stdout);
          say_error(stdout);
        }
      else if (listcis != NULL)
        {
	  writeline(listcis);
	  if (error_type != NULL)  say_error(listcis);
        }
      if (xaddr & 1) xaddr++;  /* Round up to boundary */
      if (object[0] & 1) objbyte(0); /* ditto. */
      error_type = 0;
      if (pass==2)  code_output();
      if (error_type) say_error(stdout);
      baddr[output] = baddr[output] + object[0] + xaddr;
      seclen[output] += object[0];
      if (relorigin) seclen[output] +=  xaddr;
    }
  fclose(srccis);
}


int chop(o, c)           /* check presence of case in opts */
{ 
  return 0;
}

void list_externs(int x, char *title) 
{
  int h;
  if (objcis == NULL) return;
  fprintf(objcis, title);
  for (h=0; h<HASHSIZE; h++)
    {
      struct stnode *p = stab_start[h];
      while (p)
	{ 
	  if (p->type == x)
	    { 
	      struct locchain *q = p->schain;
	      if (q)  
		{
		  fprintf(objcis, ">%s ", p->textual);
		  while (q)
		    {
		      fprintf(objcis, "%08X ", q->val);
		      q = q->next;
		    }
		  fprintf(objcis, "\n");
		}
	    }
	  p = p->stnext;
	}
    }
}


int main(int argc, char **argv)
{  
  int lenfirst;
  int h;
  /* printf("Slider 68 series assemblers.  cbg. \n");
   */
  stab_start = (struct stnode **) malloc(HASHSIZE * 4);
  for (h=0; h<HASHSIZE; h++) stab_start[h] = NULL; 
  initspecific();
  if (argc < 2) 
    {
      printf("Usnix is:  a68 -rom srcfile objfile listfile symboldumpfile\n");
      exit(1);
    }
  if (strcmp(argv[1], "-rom")==0)
    {
      readonlyopt = 1;
      argv++; argc--;
    }
  else readonlyopt = 0;
  jobname = argv[1];    
  /**====================== */
  /* FIRST PASS */
  /**====================== */
  pass = 1;
  objcis = NULL;
  if (fopen(argv[1], "r")==0)
    {
      printf("as: cant open src file ");
      perror(argv[1]);
      exit(1);
    }
  /* Module file */
  if (argc > 2)
    { 
      objcis = fopen(argv[2], "w");
      if (objcis == NULL)
	{
	  perror(argv[2]);
	  exit(1);
	}
      fprintf(objcis, "$$ CBG %s assembler job: %s\n", assembler, jobname);
    }
  listcis = NULL;
  
  
  perform_pass();
  lenfirst = baddr[O_TEXT];
  
  /**====================== */
  /* 2nd   PASS */
  /**====================== */
  if (seclen[0] + seclen[1] > 0)
    {
      pass = 2;
      
      if ((objcis != NULL))
	{
	  if (seclen[O_BSS]) fprintf(objcis, "-BL %08X\n", seclen[O_BSS]);     
	  fprintf(objcis, "-T%08X\n", seclen[O_TEXT]);
	  
	}
      if (argc > 3)
        { 
	  if (strcmp(argv[3], "=") == 0) listcis = stdout;
           else listcis = fopen(argv[3], "w");
	  if (listcis == NULL)
	    {
	      perror(argv[3]);
	      exit(1);
	    }
	  fprintf(listcis, "** CBG %s assembler job: %s\n", assembler, jobname);
	}
      perform_pass();
      list_externs(RELOCATABLE, "\n-R\n");
      list_externs(EXTERN, "\n-E\n");
      if (lenfirst != baddr[O_TEXT])
        {
	  printf("as: %s unequal passes %x  %x \n", argv[1], lenfirst, baddr[O_TEXT]);
	  errors = errors + 1;
        }
    } /* endif textlen > 0 */

  if (listcis != NULL)
    {
      fprintf(listcis, "\n\nERRORS = %i,   SRC CARDS READ = %i\n\n",
	      errors, cards);
      if (listcis != stdout) fclose(listcis);
    }

  /* dump symbols optionally
   */
  if (argc > 4)
    {
      FILE * sss = fopen(argv[4], "w");
      printf("Dumping symbols\n");
      if (listcis == NULL)
	{
	  perror(argv[4]);
	  exit(1);
	}
      for (h=0; h<HASHSIZE; h++)
	{
	  struct stnode *p = stab_start[h];
	  while (p)
	    {
	      fprintf(sss, "%08X %s\n", p->value, p->textual);
	      p = p->stnext;
	    }
	}
      fclose(sss);
    }
  

  /* Close module */
  if (objcis != NULL)
    {
      fprintf(objcis, "-%c \n", (errors>0) ? 'X': '-');
      fclose(objcis);
      if (errors) remove(argv[2]);
    }

  /* if (errors == 0) printf("Successful completion.\n");
   */
  return errors;
}



void get()
{
  if (oldcis)
    {
      printf("Nested GETs not supported.\n");
      exit(1);
    }
  oldcis = srccis;
  oldline   = line;
  oldsrc = currentsrc;
  {
    char *n = (char *) malloc(80);
    int i = 0;
    ignblank();
    while (ech != (char) 0 && ech != ';')
      {
	n[i++] = toupper(ech);
	nextch(1, 1);
      }
    n[i++] = (char) 0;
    srccis = fopen(n, "r");
    if (srccis == NULL)
      { 
	printf("Cant open included file '%s'.\n", n);
	exit(1);
      }
    /*   if (pass==1) printf("Reading: %s\n", n);
     */
    currentsrc = n;
  }
} 

/* end of a6gen .c */ /* (C) 1990, 91 DJ GREAVES. */


 
