
/* 68K specific code. 
 *
 * Tenison Tech Motorola 68K Assembler
 * (C) 1990 TenisonTech
 * This version not for commercial use.
 * License available from info@tenisontech.com
 *
 * No liability is accepted any errors in this material or for any loss.
 *
 *
 */

#include <stdio.h>
#include <ctype.h>
#include "a6hdr.h"

int mode68000 = 0; /* Flag to force 68000 instructions only */

extern int readonlyopt;
extern int output;
extern char chcomma();
extern int eval();
extern char nextch();
void bulk();
extern void error();
extern int aschex();
extern void objword();
extern void objbyte();
extern int object[];
extern void symbnote();
extern int pass, xaddr, equate, baddr[], relorigin, error_type, tab, imed, running;
extern int equateval;
extern char *token, inst[], lable[];
extern char ech;
extern struct stnode *stp;
extern struct stnode *wordnode;

extern FILE *objcis;
extern FILE *listcis;

int width, ovalue;
char tchar;

struct parseitem 
{
  int reg, mod, pval, index;
  struct stnode *symb;
} ps0, ps1;

char *nm[26];


void copy()
{  
  int m = ps0.mod;
  int r = ps0.reg;
  int v = ps0.pval;
  struct stnode *w = ps0.symb;
  ps0.mod = ps1.mod;
  ps0.reg = ps1.reg;
  ps0.pval = ps1.pval;
  ps0.symb = ps1.symb;
  ps1.mod = m;
  ps1.reg = r;
  ps1.pval = v;
  ps1.symb = w;
}

int getnum4(char *pos)
{
  int r=0;
  int i;
  for (i=0; i<4; i++)  r = (r << 4) + aschex(*(pos++));
  return r;
}


/* routine looks up nemonics (68K version). */
void lok()
{
  char xch = toupper(inst[tab++]);
  char *pos;
  if (xch == (char) 0) return;       /* blank line */
  if (xch == ';')  return;           /* comment */
  if ((xch<'A') || (xch>'Z'))  {  error("Bad char in nemonic");  return;  }
  pos = nm[(xch-'@')];          /* string of possibilities */
  while (*pos)
    {
      int iop = tab;
      
      while (toupper(inst[iop]) == *pos) // compare tail of strings
	{
	  iop = iop + 1;
	  pos++;
	}
      if (*pos == '-')               /* string matched */
	{
	  tab = iop;                 /* next input char */
	  pos ++;     
	  tchar = *(pos++);
	  ovalue = getnum4(pos);
	  bulk();
	  return;
	}
      
      while ((*pos) && (*pos != ' ')) pos++;
      while (*pos == ' ') pos++;
    }
  error("not recognized", inst);
}


void xwords(struct parseitem *vip)
{ 
  struct stnode *x = vip->symb;
  /*     (listcis, "xwords vip %x   pval %x\n", vip, vip->pval);
   */
  switch (vip->mod)
  {
  case 7:
     switch (vip->reg)
     {
     case 0: case 5:              /* short absolute or index+disp */
        if (x) error("unacceptable relocation (5)"); 
        objword(vip->pval);
        return;

     case 1:                      /* long absolute */
        if (x)  symbnote(x, baddr[output] + object[0]);
        objword((vip->pval)>>16);
        objword(vip->pval);
        return;

     case 2:                      /* PC with 16 bit displacement */
        objword(vip->pval - baddr[output] - 2);
        return;

     case 4:                      /* immediate */
        if (width==2)
        {  if (x)  symbnote(x, baddr[output] + object[0]);
            objword((vip->pval)>>16);
        }
        else if (x) error("unacceptable relocation (4)"); 
        objword(vip->pval);
        return;
      }

    case 5:
    {
      if (x) 
      {
        error("cannot relocate a 16 bit index");
      }
      objword(vip->pval);    /* PC or Ax relative */
      return;
    }

    
    case 6:   /* We support data register indirect only, at the moment */
      if (vip->pval)
      {
        /* Indexed non-indirect with base suppress, longword, scale=1 */
        objword(0x09A0 + (vip->reg<<12));
        objword(vip->pval);
      }
      else
      {
        /* Indexed non-indirect with base suppress, longword, scale=1 */
        objword(0x0990 + (vip->reg<<12));
      }
      return;  
  }    
}

void specified()                    /* handle _W _B _L specifiers */
{  
  char t = nextch(1, 1);
  nextch(1, 1);
  switch (t)
    {  case 'B': width = 0 ;
    return;
    case 'W': width = 1 ;
      return;
    case 'L': width = 2 ;
      return;
    default: error("Dot suffix size spec bad");
    }
}


int template(pat) char *pat;
{
  int i = tab-1;
  char tc = *pat;
  while(tc)
    {
      if (tc=='_') ps0.reg = inst[i] & 7;       /* address register chosen*/
      else if (toupper(inst[i]) != tc)  return 0;
      i = i + 1;
      tc = *(++pat);
    }
  tab = i;                         /* adopt template */
  nextch(1, 1);
  return 1;
}


void parse()
{
  char loka = inst[tab];           /* look ahead char */
  ps0.index = 0;
  imed = 0;
  ps0.pval = 0;
  if (ech == 'D')                 /* data reg direct */
  {
    if (loka  >=  '0' && loka  <= '7') 
    {  ps0.reg = nextch(1, 1) & 7;
           ps0.mod = 0;
           nextch(1, 1);
           return;
    }
  }

  if (ech == 'A')                 /* address register */
  {
    if (loka  >=  '0' && loka  <= '7')
    {  ps0.reg = nextch(1, 1) & 7;
      ps0.mod = 1;
      nextch(1, 1);
      return;
     }
  }

  if (template("(A_)+"))         /* post increment indirect */
  {  ps0.mod = 3;
        /* reg is set by template */
        return;
  }

  else if (template("-(A_)"))         /* pre decrement indirect */
     {  ps0.mod = 4;
        /* reg is set by template */
        return;
      }

   else if (template("USP"))         /* user stack pointer      */
   {
     ps0.mod = -2;
     return;
   }

   else if (template("(A_)"))          /* a reg indirect */
   { 
     ps0.mod = 2;
     /* reg is set by template */
     return;
   }

   else if (template("(D_)"))           /* D reg indirect (68020 and above) */
   {
     if (mode68000) error("using data reg for index");
     ps0.mod = 6;
     return;
   }

   else if ((ech == '#') || (ech=='='))                 /* immediate data */
   { 
     nextch(1, 1);
     ps0.pval = eval(&ps0.symb);
     /* fprintf(listcis, "eval %x ", ps0.pval);
      */
     
     ps0.mod = 7;
     ps0.reg = 4;
     imed = 1;
     return;
   }

  /* default to an expression */
  ps0.pval = eval(&ps0.symb);      /* get number */
  /* fprintf(listcis, "eval %x ", ps0.pval);
   */
  if (ech=='(')
    {
      nextch(1, 1);
      if (template("A_)"))
	{ 
	  ps0.mod = 5;
	  return;
	}
      if (template("D_)"))
	{ 
	  if (mode68000) error("using data reg for index");
	  ps0.mod = 6;
	  return;
	}
      else if (template("PC)"))
	{
	  ps0.mod = 7;
	  ps0.reg = 2;
	  return;
	}
    }
  if (imed)                    /* implicit immediate detected */
    {
      ps0.mod = 7;
      ps0.reg = 4;
    }
  else
    {
      ps0.mod = 7;
      if ((ps0.pval & 0xFFFF0000) || (ps0.symb)) ps0.reg = 1;
      else ps0.reg = 0;
      /* long and short */
    }
}

/* int parse()
//  {  parse1();
//  if (pass==2)
//   {
//       fprintf(listcis, "pval=%x \n", ps0.pval);
//       fprintf(listcis, "imed=%i ovalue==%04X \n", imed, ovalue);
//   }
//  }
*/

int getdreg()                 /* get a data register */
{
  int x;
  if (ech!='D')  error("Data register required.");
  x = nextch(1, 1) - '0';
  if (x<0 || x>7)  error("Register not 0 through 7.");
  nextch(1, 1);
  return x;
}

int getareg()                 /* get an address reg */
{
  int x;
  if (ech != 'A')  error("Areg expected.");
  x = nextch(1, 1) - '0';
  if (x<0 || x>7)  error("Register not 0 through 7.");
  nextch(1, 1);
  return x;
}

void check_alt(cev)                 /* check ALTERABLE */
{
  if (imed)  error("Immediate addressing not allowed.");
  /* if (cev->mod == 7)  if (cev->reg >= 2)  error("XX") */
}


void check_dalt(struct parseitem *cev)   /* check DATA & ALTERABLE */
{
  if (cev->mod == 1)  error("Data alterable mode expected.");
  check_alt(cev);
}

void check_data(cev) struct parseitem *cev;   /* check DATA */
{
  if (cev->mod == 1)  error("Data mode expected.");
}

void check_areg(m)
{
  if (m != 1)  error("Address reg needed.");
}

void initspecific()
{
  nm[01] = "SL-RE100 SR-RE000 DDI-30600 NDI-30200 DDQ-q5000 "
    "ND-U0004 CON-D0005 DD-U0000 LIGN-D0009";
  nm[02] = "CC-B6400 CS-B6500 EQ-B6700 GT-B6E00 HI-B6200 \
LE-B6F00 LS-B6300 LT-B6D00 MI-B6B00 NE-B6600 PL-B6A00 \
VC-B6800 VS-B6900 RA-B6000 SR-B6100 TST-J0800 SET-J08C0 CLR-J0880 \
CHG-J0140 GE-B6C00 SS-D0008";
  nm[03] = "LR-E4200 MPA.L-pB1C0 MPI-30C00 MP-PB000 -B6100";
  nm[04] = "S.B-Z0000 S.W-Z0001 S.L-Z0002 ATA-D0004 C.L-D0005 "
    "IVS-C81C0 IVU-C80C0 BEQ-b57C8 BRA-b51C8 C.B-D0004";
  nm[05] = "ND-D0001 QU-Z0004 XT.B-W4880 XT.W-W48C0 XTB.L-W49C0 \
XG-XC100 ORI-30A00 OR-QB100 XTERN-Z0005";
  nm[6] = "";
  nm[7] = "LOBAL-z0009";
  nm[8] = "";
  nm[9] = "";
  nm[10] = "MP-F4EC0 SR-F4E80";
  nm[11] = "";
  nm[12] = "SL-RE108 SR-RE008 EA-p41C0";
  nm[13] = "OVEC-!4E7A OVE-U0002 ULS-CC1C0 ULU-CC0C0";
  nm[14] = "OP-I4E71 BCD-F4800 EG-E4400 EGX-E4000 OT-E4600";
  nm[15] = "RGR-D0006 RG-D0002 RI-30000 R-U0003";
  nm[16] = "EA-F4840";
  nm[17] = "";
  nm[18] = "TR-I4E77 TS-I4E75 TE-I4E73 OL-RE118 OR-RE018 OXL-RE110 OXR-RE010";
  nm[19] = "TOP-S4E72 WAP-W4840 UBI-30400 UBQ-q5100 NE-f56C0 GT-f5EC0 LT-f5DC0 EQ-f57C0 LE-f5FC0 GE-f5CC0 UB-U0001";
  nm[20] = "RAPV-I4E76 RAP-T4E40 AS-F4AC0 ST-E4A00 EXT-D0003";
  nm[21] = "NLK-A4E58";
  nm[22] = "";
  nm[23] = "";
  nm[24] = "";
  nm[25] = "";
  nm[26] = "";
}

int modetab[] = {3, 3, 7};
int movetab[] = {1, 3, 2};
int movecode[] = { 0xD000, 0x9000, 0, 0x8000, 0xC000 };

void bulk()
{
  nextch(1, 1);                        /* preload one character */
  width = 0;                       /* default data size is byte */
  if (ech=='.')  specified();
  switch (tchar)
    {
    default:
      printf("a68k Internal error, bulk '%c'\n", tchar);
      exit(1);
      
    case 'A':                        /* opcode + address register */
      objword(getareg()+ovalue);
      return;
    case 'B':                        /* rel hops */
      {  int delta = eval(&ps0.symb) - baddr[output] - 2;
      if (error_type > 0)        /* assume first pass forward ref */
	{  objword(ovalue);
	objword(0);
	return;
	}
      if (delta>0 || delta<-128)   /* forward ref or big -ve hop */
	{  objword(ovalue);        /* then use word format */
	objword(delta);
	return;
	}
      objword((delta & 255) + ovalue);  /* short form */
      return;
      }
    case 'b':                        /* dbra instruction */
      {  int dreg = getdreg();
      int delta;
      chcomma();
      delta = eval(&ps0.symb) - baddr[output] - 2;
      objword(ovalue + dreg);
      objword(delta);
      return;
      }
    case 'C':                        /* multiply and divide */
      {  int dreg = getdreg();
      chcomma();
      parse();
      objword(ovalue + (dreg<<9) + (ps0.mod<<3) + ps0.reg);
      xwords(&ps0);
      check_data(&ps0);
      return;
      }
    
    case 'D':                        /* assembler directives */
      { int byte_flag;
      switch (ovalue)
	{  case 1: running = 0; /* end directive */
	return;
	case 2: baddr[output] = eval(&ps0.symb); /* origin directive */
	  relorigin = 0;
	  equate = 1;
	  equateval = baddr[output];
	  return;
	  
	case 8: 
	  output = O_BSS;
	  return;
	  
	case 3: 
	  output = O_TEXT;
	  return;
	  
	  
	case 6: baddr[output] = eval(&ps0.symb); /* originr directive */
	  relorigin = 1;
	  equate = 1;
	  equateval = baddr[output];
	  return;
	case 4:                  /* data directive */
	  byte_flag = 0;
	  while(1)
	    {  if (ech=='\'')  /* quoted string */
	      {  nextch(0, 0);
	      while (ech !='\'')
		{  objbyte(ech);
		ech = nextch(0, 0);
		if (ech == 0)  
		  {  error("End of line in string");
		  break;
		  }
		}
	      nextch(1, 1);
	      }
	    else objbyte(eval(&ps0.symb)); /* data expression */
	    if (ech==0 || ech==';') return;
	    chcomma();
	    } 
	  
	case 5:                  /* ACON directive */
	  while(1)
	    {  int a = eval(&ps0.symb);
	    if (ps0.symb)  symbnote(ps0.symb, baddr[output] + object[0]);
	    objword(a>>16) ; objword(a & 0xFFFF);
	    if (ech==0 || ech==';')  return;
	    chcomma();
	    } 
	case 9:       /* Align directive */
	  {
	    int a = eval(&ps0.symb);
	    a = a - 1;
	    while ((baddr[output]+xaddr) & a) xaddr++;
	    return;
	  }                           
	}
      }
    case 'E':                        /* such as CLR and NOT */
      {  parse();                     /* one effective address */
      objword(ovalue + (width<<6) + (ps0.mod<<3) + ps0.reg);
      xwords(&ps0);
      check_dalt(& ps0);
      return;
      }
    case 'F':  case 'f':             /* such as JSR and TAS */
      {  parse();                      /* same as 'E' but no width */
      /*        if (tchar!='f') check_control(ps0); REINSTATE @@@@ */
      if ((ps0.mod == 7) && (ps0.reg==0))  ps0.reg = 1;
      objword(ovalue + (ps0.mod<<3) + ps0.reg);
      xwords(&ps0);
      return;
      }
    case 'I':                        /* pramless opcodes */
      objword(ovalue);
      return;
      
    case 'J':                         /* bit operations */
      {  int bit = eval(&ps0.symb);
      chcomma();
      parse();
      objword(ovalue + (ps0.mod<<3) + ps0.reg);
      objword(bit);
      xwords(&ps0);
      return;
      }
    case 'p':                         /* LEA, cmpa_l */
      parse();
      chcomma();
      width = 2;
      {  int rrr = getareg();
      objword(ovalue +(rrr<<9) + (ps0.mod<<3) + ps0.reg);
      xwords(&ps0);
      return;
      }
    case 'P':                         /* <ea>,Dn or <ea>,An (CMP CMPA) */
      parse();
      chcomma();
      if (ech=='A')             /* Aregister type */
	{
	  int rrr = getareg();
	  int opmode = modetab[width];
	  objword(ovalue+(rrr<<9)+(opmode<<6)+(ps0.mod<<3)+ps0.reg);
	}
      else                         /* d reg type */
	{
	  int rrr = getdreg();
	  objword(ovalue+(rrr<<9)+(width<<6)+(ps0.mod<<3)+ps0.reg);
	}
      xwords(&ps0);
      return;
      
    case 'Q':                         /* Dn, <ea> */
     { 
       int rrr = getdreg();
       chcomma();
       parse();
       check_dalt(&ps0);
       objword(ovalue + (rrr<<9) + (width<<6) + (ps0.mod<<3) + ps0.reg);
       xwords(&ps0);
       return;
     }
    
   case 'q':                         /* addq, subq  imm,dadr */
     {
       int dval;
       ps0.symb = NULL;
       dval = eval(&ps0.symb);
       if (dval<1 || dval>8)  error("quick value not between 1 and 8");
       dval = dval & 7;
       
       chcomma();
       parse();
       check_alt(&ps0);
       objword(ovalue + (dval<<9) + (width<<6) + (ps0.mod<<3) + ps0.reg);
       xwords(&ps0);
       return;
     }
   
    case 'R':                         /* rotate instructions */
      {
	int regno;
        int count;
        parse();
	
        if (ps0.mod == 0)
          {
	    ovalue = ovalue + 0x20;  /* Shift value in a register */
	    count = ps0.reg;
	  }
        else
          {
	    count = ps0.pval;
	    if (count<1 || count>8)  error("count not between 1 and 8");
	  }
        count = count & 7;
        chcomma();
        regno = getdreg();           /* data register */
        objword(ovalue+regno+(width<<6)+(count<<9));
        return;
      }
    case 'S':                        /* sixteen bit immediate data */
      objword(ovalue);
      objword(eval(&ps0.symb));
      return;
      
    case 'T':                        /* TRAP # instruction */
      {
	int i = eval(&ps0.symb) & 15;
	objword(ovalue+i);
        return;
      }
    
    /* 0==ADD  1==SUB  2==MOVE  3==OR 4==AND */
    case 'U':                        /* dn, ea or ea,dn */
      {  
	int swap = 0;
        parse();                     /* source */
        copy();                      /* to 1_ */
        chcomma();
        parse();                     /* Destination to 0_ */
	
        if (ovalue == 2 && ps0.mod == -2)              /* move to usp */
	  {
	    objword(0x4E60 + ps1.reg);  /* reg1 should be an a reg */
	    check_areg(ps1.mod);
            return;
	  }
	
        if (ovalue == 2 && ps1.mod == -2)              /* move from usp */
	  {
	    objword(0x4E68 + ps0.reg);  /* reg0 should be an a reg */
	    check_areg(ps0.mod);
            return;
	  }

        if (width==0 && ps0.mod==1)  width=1;  /* assign byte to A reg */
        if (ovalue >= 3) check_dalt(&ps0); /* ANDA and ORA do not exist */
        else check_alt(&ps0);           /* check out destination */
        if (ovalue==2)              /* move instruction */
         {
	   int opmode = movetab[width];
	   if (width==0 && ps1.mod==7 && ps1.reg==4 && ps0.mod==0)
             {
	       objword(0x7000 + (ps0.reg<<9) + (ps1.pval & 255));
	       return;                            /* Moveq inst */
	     }
	   objword((opmode<<12) + (ps0.reg<<9) + (ps0.mod<<6) + (ps1.mod<<3) +ps1.reg);
	   xwords(& ps1);
	   xwords(& ps0);
	   return;
	 }
	
	/* Support implicit adda instruction */
        if ((ps0.mod == 1) && (ovalue==0))
	  objword(0xD1C0 + (ps0.reg<<9) + (ps1.mod << 3) + ps1.reg);
        else
	  /* For other cases, one operand must be a data register */
	  { 
	    if (ps0.mod)  {  copy(); swap = 1;  }
	    if (ps0.mod)  error("One operand not a D reg");
	    ovalue = movecode[ovalue];
	    objword(ovalue + (ps0.reg<<9) + (swap<<8) + (width<<6) + (ps1.mod<<3) + ps1.reg);
	  }
        xwords(& ps1);
        return;
      }

    case 'W':                        /* opcode + data register */
        objword(getdreg()+ovalue);
        return;

    case 'X':                        /* exg register instruction */
     {  int opmode;
        int rx;
        int ry;
        if (ech=='A') 
         {  opmode = 9;               /* exchange A A */
            rx = getareg();
            chcomma();
            ry = getareg();
          }
        else
         {  rx = getdreg();
            chcomma();
            if (ech=='A') 
             {  opmode = 17;         /* exchange D A */
                ry = getareg();
              }
            else
             {  opmode = 8;          /* exchange D D */
                ry = getdreg();
              }
          }
        objword(ovalue + (rx<<9) + (opmode<<3) + ry);
        return;
      }

    case 'Z':               /* Storage allocation directives */
     {  struct stnode *symbol = wordnode;
        if (symbol->type == DMTO) return;  /* dmto condition */

        switch (ovalue)
         {  case 0: 
                ps0.symb = NULL;
                equateval = (eval(&ps0.symb)+1) & 0xFFFFE; /* Round up to word */
                xaddr = equateval;
                equate = 1;
                if (readonlyopt) error("Not readonly");
                return;

            case 1:
                ps0.symb = NULL;
                equateval = eval(&ps0.symb);
                xaddr = equateval << 1;
                equate = 1;
                if (readonlyopt) error("Not readonly");
                return;
            case 2:
                ps0.symb = NULL;
                equateval = eval(&ps0.symb);
                xaddr = equateval << 2;
                equate = 1;
                if (readonlyopt) error("Not readonly");
                return;

            case 4:   /* Equate directive */
                ps0.symb = NULL;
                equateval = eval(&ps0.symb);
                symbol->value = equateval;
                symbol->type = (ps0.symb) ? ps0.symb->type : ABSOLUTE;
                symbol->equatef = 1;
                equate = 1;
                return;

            case 5: /* EXTERN directive - not handled here*/
             return;
          }
      }

    case 'z':                        /* GLOBAL directive */
      if (pass != 1)  return;

      if (objcis && output == O_TEXT && wordnode)
      {
        fprintf(objcis, "-G %s %c%08X\n", 
           wordnode->textual, 
           (wordnode->type == RELOCATABLE) ? 'r':' ',  wordnode->value);
      }
      return;

    case '3':                        /* immediate addressing explicit */
     {
        int dwidth;
        parse();  copy();
        dwidth = width;
        chcomma();
        parse();
        width = dwidth;
        check_dalt(& ps0);
        objword(ovalue + (width<<6) + (ps0.mod<<3) + ps0.reg);
        ps1.reg = 4; ps1.mod = 7;
        xwords(& ps1);
        xwords(& ps0);
        return;
      }

   case '!':                        /* movec instruction */
     { 
       parse(); chcomma(); copy(); parse();
       if (ps0.mod == 0 || ps0.mod == 1)        /* move from a c-reg */
         {  
	   objword(ovalue);
	   objword((ps0.mod<<15) + (ps0.reg << 12) + ps1.pval);
         }
       else                              // to c-reg from a or d reg
         {
	   objword(ovalue+1);
	   objword((ps1.mod<<15) + (ps1.reg << 12) + ps0.pval);
         }
       return;
     }
    }
}

/* End of spec68k.c */

